# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Source of truth for the dse_do_utils version.
The versions in `setup.py` and `/docs/source/conf.py` are automatically populated from here.

Best practice to keep version here, in a separate file.
See https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
"""

__version__ = "0.1.2.1b0"
